import $ from '../../core/renderer';
import CollectionWidget from './ui.collection_widget.edit';
import { extend } from '../../core/utils/extend';
import arrayUtils from '../../data/array_utils';
import { each } from '../../core/utils/iterator';
import { keysEqual } from '../../data/utils';
import { when } from '../../core/utils/deferred';
import { findChanges } from '../../core/utils/array_compare';
import { insertElement } from '../../core/dom_adapter';
import { noop } from '../../core/utils/common';

const PRIVATE_KEY_FIELD = '__dx_key__';

export default CollectionWidget.inherit({
    _getDefaultOptions: function() {
        return extend(this.callBase(), {
            repaintChangesOnly: false
        });
    },

    ctor: function() {
        this.callBase.apply(this, arguments);

        this._customizeStoreLoadOptions = (e) => {
            const dataSource = this._dataSource;
            if(dataSource && !dataSource.isLoaded()) {
                this._correctionIndex = 0;
            }
            if(this._correctionIndex && e.storeLoadOptions) {
                e.storeLoadOptions.skip += this._correctionIndex;
            }
        },

        this._dataSource && this._dataSource.on('customizeStoreLoadOptions', this._customizeStoreLoadOptions);
    },

    reload: function() {
        this._correctionIndex = 0;
    },

    _init: function() {
        this.callBase();
        this._refreshItemsCache();
        this._correctionIndex = 0;
    },

    _findItemElementByKey: function(key) {
        let result = $();
        const keyExpr = this.key();
        this.itemElements().each((_, item) => {
            const $item = $(item);
            const itemData = this._getItemData($item);
            if(keyExpr ? keysEqual(keyExpr, this.keyOf(itemData), key) : this._isItemEquals(itemData, key)) {
                result = $item;
                return false;
            }
        });
        return result;
    },

    _dataSourceChangedHandler: function(newItems, e) {
        if(e?.changes) {
            this._modifyByChanges(e.changes);
        } else {
            this.callBase(newItems, e);
            this._refreshItemsCache();
        }
    },

    _isItemEquals: function(item1, item2) {
        if(item1 && item1[PRIVATE_KEY_FIELD]) {
            item1 = item1.data;
        }

        try {
            return JSON.stringify(item1) === JSON.stringify(item2);
        } catch(e) {
            return item1 === item2;
        }
    },

    _isItemStrictEquals: function(item1, item2) {
        return this._isItemEquals(item1, item2);
    },

    _shouldAddNewGroup: function(changes, items) {
        let result = false;
        if(this.option('grouped')) {
            each(changes, (i, change) => {
                if(change.type === 'insert') {
                    result = true;
                    each(items, (_, item) => {
                        if(change.data.key === item.key) {
                            result = false;
                            return false;
                        }
                    });
                }
            });
        }

        return result;
    },

    _partialRefresh: function() {
        if(this.option('repaintChangesOnly')) {
            const keyOf = (data) => {
                if(data && data[PRIVATE_KEY_FIELD] !== undefined) {
                    return data[PRIVATE_KEY_FIELD];
                }
                return this.keyOf(data);
            };
            const result = findChanges(this._itemsCache, this._editStrategy.itemsGetter(), keyOf, this._isItemStrictEquals.bind(this));
            if(result && this._itemsCache.length && !this._shouldAddNewGroup(result, this._itemsCache)) {
                this._modifyByChanges(result, true);
                this._renderEmptyMessage();
                return true;
            } else {
                this._refreshItemsCache();
            }
        }
        return false;
    },

    _refreshItemsCache: function() {
        if(this.option('repaintChangesOnly')) {
            const items = this._editStrategy.itemsGetter();
            try {
                this._itemsCache = extend(true, [], items);
                if(!this.key()) {
                    this._itemsCache = this._itemsCache.map((itemCache, index) => {
                        return {
                            [PRIVATE_KEY_FIELD]: items[index],
                            data: itemCache
                        };
                    });
                }
            } catch(e) {
                this._itemsCache = extend([], items);
            }
        }
    },

    _dispose: function() {
        this._dataSource && this._dataSource.off('customizeStoreLoadOptions', this._customizeStoreLoadOptions);
        this.callBase();
    },

    _updateByChange: function(keyInfo, items, change, isPartialRefresh) {
        if(isPartialRefresh) {
            this._renderItem(change.index, change.data, null, this._findItemElementByKey(change.key));
        } else {
            const changedItem = items[arrayUtils.indexByKey(keyInfo, items, change.key)];
            if(changedItem) {
                arrayUtils.update(keyInfo, items, change.key, change.data).done(() => {
                    this._renderItem(items.indexOf(changedItem), changedItem, null, this._findItemElementByKey(change.key));
                });
            }
        }
    },

    _insertByChange: function(keyInfo, items, change, isPartialRefresh) {
        when(isPartialRefresh || arrayUtils.insert(keyInfo, items, change.data, change.index)).done(() => {
            this._beforeItemElementInserted(change);

            const $itemContainer = this._getItemContainer(change.data);

            this._renderItem(change.index ?? items.length, change.data, $itemContainer);

            this._afterItemElementInserted();
            this._correctionIndex++;
        });
    },

    _getItemContainer: function(changeData) {
        return this._itemContainer();
    },

    _updateSelectionAfterRemoveByChange: function(removeIndex) {
        const selectedIndex = this.option('selectedIndex');

        if(selectedIndex > removeIndex) {
            this.option('selectedIndex', selectedIndex - 1);
        } else if(selectedIndex === removeIndex && this.option('selectedItems').length === 1) {
            this.option('selectedItems', []);
        } else {
            this._normalizeSelectedItems();
        }
    },

    _beforeItemElementInserted: function(change) {
        const selectedIndex = this.option('selectedIndex');

        if(change.index <= selectedIndex) {
            this.option('selectedIndex', selectedIndex + 1);
        }
    },

    _afterItemElementInserted: noop,

    _removeByChange: function(keyInfo, items, change, isPartialRefresh) {
        const index = isPartialRefresh ? change.index : arrayUtils.indexByKey(keyInfo, items, change.key);
        const removedItem = isPartialRefresh ? change.oldItem : items[index];
        if(removedItem) {
            const $removedItemElement = this._findItemElementByKey(change.key);
            const deletedActionArgs = this._extendActionArgs($removedItemElement);

            this._waitDeletingPrepare($removedItemElement).done(()=>{
                if(isPartialRefresh) {
                    this._updateIndicesAfterIndex(index - 1);
                    this._afterItemElementDeleted($removedItemElement, deletedActionArgs);
                    this._updateSelectionAfterRemoveByChange(index);
                } else {
                    this._deleteItemElementByIndex(index);
                    this._afterItemElementDeleted($removedItemElement, deletedActionArgs);
                }
            });

            this._correctionIndex--;
        }
    },

    _modifyByChanges: function(changes, isPartialRefresh) {
        const items = this._editStrategy.itemsGetter();
        const keyInfo = { key: this.key.bind(this), keyOf: this.keyOf.bind(this) };
        const dataSource = this._dataSource;
        const paginate = dataSource && dataSource.paginate();
        const group = dataSource && dataSource.group();

        if(paginate || group) {
            changes = changes.filter(item => item.type !== 'insert' || item.index !== undefined);
        }

        changes.forEach(change => this[`_${change.type}ByChange`](keyInfo, items, change, isPartialRefresh));
        this._renderedItemsCount = items.length;
        this._refreshItemsCache();
        this._fireContentReadyAction();
    },

    _appendItemToContainer: function($container, $itemFrame, index) {
        const nextSiblingElement = $container.children(this._itemSelector()).get(index);
        insertElement($container.get(0), $itemFrame.get(0), nextSiblingElement);
    },

    _optionChanged: function(args) {
        switch(args.name) {
            case 'items': {
                const isItemsUpdated = this._partialRefresh(args.value);
                if(!isItemsUpdated) {
                    this.callBase(args);
                }
                break;
            }
            case 'dataSource':
                if(!this.option('repaintChangesOnly') || !args.value) {
                    this.option('items', []);
                }

                this.callBase(args);
                break;
            case 'repaintChangesOnly':
                break;
            default:
                this.callBase(args);
        }
    }
});
